# ----------------------------------------------------------------------------------------------------
# IBM Confidential
# OCO Source Materials
# 5900-A3Q, 5737-H76
# Copyright IBM Corp. 2021
# The source code for this program is not published or other-wise divested of its trade 
# secrets, irrespective of what has been deposited with the U.S.Copyright Office.
# ----------------------------------------------------------------------------------------------------

import copy
import json
import time
try:
    from pyspark.sql.dataframe import DataFrame
except ImportError as ie:
    pass

from ibm_wos_utils.fairness.batch.utils.constants import CLASS_PROBABILITY_MODELING_ROLE, FEATURE_MODELING_ROLE, META_FIELD_MODELING_ROLE, PREDICTION_MODELING_ROLE, PROBABILITY_MODELING_ROLE, RECORD_ID_MODELING_ROLE, REGRESSION_MODEL_TYPE, SUPPORTED_STORAGE_TYPES, TIMESTAMP_MODELING_ROLE
from ibm_wos_utils.fairness.batch.utils.batch_utils import BatchUtils
from ibm_wos_utils.fairness.batch.utils.perturbation_utils import PerturbationUtils
from ibm_wos_utils.fairness.batch.utils.python_util import get
from ibm_wos_utils.fairness.batch.utils.sql_utils import SQLUtils
from ibm_wos_utils.joblib.clients.scoring_client import ScoringClient
from ibm_wos_utils.joblib.jobs.aios_spark_job import AIOSBaseJob
from ibm_wos_utils.joblib.utils.db_utils import DbUtils
from ibm_wos_utils.joblib.utils.joblib_utils import JoblibUtils


class GroupBiasComputationJob(AIOSBaseJob):

    def __init__(self, arguments, job_name):
        """
        Constructor for the job class.
        :arguments: The arguments to the Spark job.
        """
        super().__init__(arguments)
        self.name = job_name

    def calculate_group_bias(self, data, inputs: dict, data_types: dict, model_type: str, total_group_counts: dict=None) -> dict:
        """
        The Spark job which calculates the disparate impact ratio and publishes the fairness metrics for the payload and their corresponding perturbed data.
        :data: The spark data frame containing the payload data. (from pyspark.sql.dataframe import DataFrame)
        :inputs: The inputs dictionary.
        :data_types: The dictionary containing data types of all the fairness attributes.
        :model_type: The model type.
        :total_group_counts: [Optional] Dictionary containing counts for each group of all fairness attributes present in the fairness configuration. This if given, optimises the DI calculation.
            Example,
                If inputs = {
                    "fairness_attributes": ["age", "gender"],
                    "majority": {
                        "age": [[18, 25]],
                        "gender": ["male"]
                    },
                    "minority": {
                        "age": [[26, 90]],
                        "gender": ["female]
                    }...
                } then,
                total_group_counts = {
                    "age": {
                        "[18, 25]": 120,
                        "[26, 90]": 180
                    },
                    "gender": {
                        "male": 220,
                        "female": 80
                    }
                }
        
        :returns: The DI dictionary.
        """
        # First calculating the disparate impact on the payload data
        di_dict = BatchUtils.calculate_di_dict(
            data, inputs, data_types, model_type, total_group_counts=total_group_counts)
        return di_dict

    def run_job(self) -> None:
        """
        The entry point method for the Spark job.
        """
        self.logger.info("Started the group bias computation job run.")

        try:
            # Reading the inputs from the argument list
            subscription = get(self.arguments, "subscription")
            monitor_instance = get(self.arguments, "monitor_instance")
            output_file_path = get(self.arguments, "output_file_path")
            service_provider = get(self.arguments, "service_provider")
            service_provider_credentials = get(self.arguments, "service_provider_credentials")

            # Reading configuration parameters from monitor instance
            perform_perturbation = get(monitor_instance, "entity.parameters.perform_perturbation", default=False)
            sample_size_percent = get(monitor_instance, "entity.parameters.sample_size_percent")
            numerical_perturb_count_per_row = get(monitor_instance, "entity.parameters.numerical_perturb_count_per_row", default=2)
            float_decimal_place_precision = get(monitor_instance, "entity.parameters.float_decimal_place_precision")
            numerical_perturb_seed = get(monitor_instance, "entity.parameters.numerical_perturb_seed")
            scoring_page_size = get(monitor_instance, "entity.parameters.scoring_page_size", default=1000)
            score_in_pages = True
            copy_measurement = False

            if self.storage_type not in SUPPORTED_STORAGE_TYPES:
                raise Exception("{} storage type is not supported. Supported storage types are {}".format(self.storage_type, SUPPORTED_STORAGE_TYPES))

            # Getting the inputs dictionary
            inputs = BatchUtils.get_inputs_from_monitor_instance(
                monitor_instance)
            
            # Adding the class label in inputs dictionary
            output_data_schema = get(subscription, "entity.asset_properties.output_data_schema")
            inputs["class_label"] = BatchUtils.get_name_with_modeling_role("prediction", output_data_schema)
            min_records = get(monitor_instance, "entity.parameters.min_records")

            # Checking if record timestamp column is present in the PL table
            timestamp_present = BatchUtils.check_if_modeling_role_present(TIMESTAMP_MODELING_ROLE, output_data_schema)
            if not timestamp_present:
                # Raise an exception as the timestamp column is mandatory because of #19570
                raise Exception("Mandatory timestamp column is not present in the schema!")

            # Getting the payload logger data source
            pl_data_source = BatchUtils.get_data_source_from_subscription(
                subscription, "payload")
            db_name = get(pl_data_source, "database_name")
            pl_table_name = get(pl_data_source, "table_name")
            pl_schema_name = get(pl_data_source, "schema_name")

            # Reading the data
            payload_start_time = time.time()
            df_spark, borrow_if_needed = self._read_data(
                subscription, monitor_instance, db_name, pl_table_name, schema_name=pl_schema_name)

            # Getting the model type and the data types of the fairness attributes
            model_type = get(subscription, "entity.asset.problem_type")
            data_types = BatchUtils.get_data_types(
                subscription, inputs["fairness_attributes"])

            di_dict = self.calculate_group_bias(df_spark, inputs, data_types, model_type)
            rows_analyzed = get(di_dict, "rows_analyzed")

            balanced_di_dict = None
            if perform_perturbation:
                payload_end_time = time.time()
                payload_time_taken = payload_end_time - payload_start_time
                if rows_analyzed == 0:
                    # No new records were read, adding flag to copy the previous measurements, if exists
                    self.logger.info("No new records were read, adding the copy measurement flag in output file.")
                    copy_measurement = True
                else:
                    # Reading stratified sample with sample size percent for the last window
                    balanced_start_time = time.time()
                    
                    # Getting the record-id column
                    record_id_column = BatchUtils.get_name_with_modeling_role(RECORD_ID_MODELING_ROLE, output_data_schema)

                    # Getting the timestamp column name
                    timestamp_column = BatchUtils.get_name_with_modeling_role(TIMESTAMP_MODELING_ROLE, output_data_schema)
                    
                    # Getting the probability column (Would be None in case of regression models)
                    probability_column = BatchUtils.get_name_with_modeling_role(PROBABILITY_MODELING_ROLE, output_data_schema)

                    # Getting all the columns
                    columns_to_map = [field["name"] for field in get(output_data_schema, "fields")]
                    columns_to_filter = [record_id_column]
                    
                    # Checking the last processed time
                    last_processed_time = get(monitor_instance, "entity.parameters.last_processed_ts")

                    # Getting the partition information
                    payload_data_source = BatchUtils.get_data_source_from_subscription(subscription, "payload")
                    partition_column = get(payload_data_source, "parameters.partition_column")
                    num_partitions = get(payload_data_source, "parameters.num_partitions")

                    # Getting the class labels and class probabilities if available (SPSS) https://github.ibm.com/aiopenscale/tracker/issues/28078
                    class_labels = BatchUtils.get_class_labels_from_prediction_metadata(output_data_schema)
                    class_probabilities = BatchUtils.get_names_with_modeling_role(CLASS_PROBABILITY_MODELING_ROLE, output_data_schema)

                    # Initialising the scoring client
                    feature_columns = BatchUtils.get_names_with_modeling_role(FEATURE_MODELING_ROLE, output_data_schema)
                    prediction_column = BatchUtils.get_name_with_modeling_role(PREDICTION_MODELING_ROLE, output_data_schema)
                    scoring_url = get(subscription, "entity.deployment.scoring_endpoint.url", default=get(subscription, "entity.deployment.url"))
                    meta_fields = BatchUtils.get_names_with_modeling_role(META_FIELD_MODELING_ROLE, output_data_schema)
                    service_type = get(service_provider, "entity.service_type")
                    prediction_data_type = BatchUtils.get_data_types(subscription, [prediction_column])[prediction_column]
                    scoring_client = ScoringClient(
                        service_type=service_type,
                        model_type=model_type,
                        scoring_url=scoring_url,
                        credentials=service_provider_credentials,
                        features=feature_columns,
                        prediction=prediction_column,
                        probability=probability_column,
                        prediction_data_type=prediction_data_type,
                        metafields=meta_fields,
                        page_size=scoring_page_size,
                        class_labels=class_labels,
                        class_probabilities=class_probabilities
                    )

                     # Getting the scoring output fields
                    scoring_output_fields = []
                    for field_schema in output_data_schema["fields"]:
                        field_name = field_schema["name"]
                        field_modelling_role = get(field_schema, "metadata.modeling_role")
                        is_deleted_field = get(field_schema, "metadata.deleted")
                        if field_modelling_role is not None and field_modelling_role in [FEATURE_MODELING_ROLE, META_FIELD_MODELING_ROLE, PREDICTION_MODELING_ROLE, PROBABILITY_MODELING_ROLE] and not is_deleted_field:
                            if "." in field_name:
                                field_name = "`{}`".format(field_name)
                            scoring_output_fields.append(field_name)

                    # Accumulating the input columns for scoring
                    input_data_schema = get(subscription, "entity.asset_properties.input_data_schema")
                    input_columns = JoblibUtils.get_columns_with_modeling_role(input_data_schema, "feature")
                    meta_columns = JoblibUtils.get_columns_with_modeling_role(input_data_schema, "meta-field")
                    input_columns.extend(meta_columns)

                    # Getting all the labels
                    all_labels = inputs["favourable_class"] + inputs["unfavourable_class"]

                    # Initialising the perturb data size counter
                    perturbed_data_size = 0

                    # Perturbing the sample read, scoring and calculating DI for balanced
                    for fairness_attribute in inputs["fairness_attributes"]:
                        # Building inputs just for current fairness attribute
                        fair_att_inputs = copy.deepcopy(inputs)
                        fair_att_inputs["fairness_attributes"] = [fairness_attribute]
                        fair_att_inputs["majority"] = {
                            fairness_attribute: inputs["majority"][fairness_attribute]
                        }
                        fair_att_inputs["minority"] = {
                            fairness_attribute: inputs["minority"][fairness_attribute]
                        }
                        fair_att_inputs["threshold"] = [inputs["threshold"][inputs["fairness_attributes"].index(fairness_attribute)]]

                        self.logger.info("Reading the stratified sample for fairness attribute {} balanced data calculation.".format(fairness_attribute))
                        balanced_sample_read_start_time = time.time()

                        # For a protected attribute, checking if correlated attributes exist. If not found, then perturbation is not performed.
                        is_protected_attribute = fairness_attribute in inputs.get("protected_attributes")
                        if is_protected_attribute and BatchUtils.should_skip_perturbation_for_prot_attr(inputs, fairness_attribute):
                            self.logger.info("No correlated attributes found for the protected attribute {}, hence skipping perturbation.".format(fairness_attribute))
                            # Update balanced DI with the same value as payload DI
                            balanced_di_dict[fairness_attribute] = di_dict.get(fairness_attribute)
                            continue

                        # If last_processed time is None, we will read all the data as it would be the first run,
                        # else, we would read from the last window
                        balanced_pl_sample_df, balanced_pl_sample_total_group_counts = BatchUtils.read_stratified_samples(
                            sample_size_percent=sample_size_percent,
                            inputs=fair_att_inputs,
                            data_types=data_types,
                            spark=self.spark,
                            location_type=self.location_type,
                            db_name=db_name,
                            table_name=pl_table_name,
                            columns_to_map=columns_to_map,
                            columns_to_filter=columns_to_filter,
                            model_type=model_type,
                            all_labels=all_labels,
                            schema_name=pl_schema_name,
                            jdbc_connection_properties=self.jdbc_connection_properties,
                            record_timestamp_column=timestamp_column,
                            start_time=last_processed_time,
                            probability_column=probability_column,
                            order_by_timestamp_desc=True,
                            partition_column=partition_column,
                            num_partitions=num_partitions
                        )
                        balanced_sample_read_end_time = time.time()
                        self.logger.info("Completed reading the stratified sample for fairness attribute {} balanced dataset calculation. Time taken {} seconds.".format(fairness_attribute, balanced_sample_read_end_time - balanced_sample_read_start_time))

                        # Filtering columns from data read other than scoring output fields
                        balanced_pl_sample_df = balanced_pl_sample_df[scoring_output_fields]
                        
                        # Starting the perturbation
                        self.logger.info("Starting the perturbation of the sample read for balanced dataset for fairness attribute {}.".format(fairness_attribute))
                        sample_pert_df = None
                        # Counts for each group in perturbed data
                        total_pert_group_counts = {
                            fairness_attribute: dict()
                        }
                        # Correlation information for protected attribute
                        correlation_info = None
                        if is_protected_attribute:
                            correlated_attributes = get(inputs, "correlated_attributes.{}".format(fairness_attribute))
                            correlated_majority = get(inputs, "correlated_majority.{}".format(fairness_attribute))
                            correlated_minority = get(inputs, "correlated_minority.{}".format(fairness_attribute))
                            correlation_info = {
                                "correlated_attributes": correlated_attributes,
                                "correlated_maj_min": correlated_majority
                            }

                        pert_start_time = time.time()
                        # Perturbing the minority groups
                        for mino in fair_att_inputs["minority"][fairness_attribute]:
                            from_group_count = balanced_pl_sample_total_group_counts[fairness_attribute][str(mino)]
                            to_groups = fair_att_inputs["majority"][fairness_attribute]
                            mino_pert_df, pert_to_groups_count = PerturbationUtils.perform_perturbation(
                                data=balanced_pl_sample_df,
                                column_to_perturb=fairness_attribute,
                                from_group=mino,
                                to_groups=to_groups,
                                spark=self.spark,
                                numerical_perturb_count_per_row=numerical_perturb_count_per_row,
                                float_decimal_place_precision=float_decimal_place_precision,
                                numerical_perturb_seed=numerical_perturb_seed,
                                from_group_count=from_group_count,
                                correlation_info=correlation_info
                            )
                            sample_pert_df = sample_pert_df.unionByName(mino_pert_df) if sample_pert_df is not None else mino_pert_df
                            for to_group in to_groups:
                                perturbed_to_group_count = pert_to_groups_count[str(to_group)]
                                perturbed_data_size += perturbed_to_group_count
                                if str(to_group) not in total_pert_group_counts[fairness_attribute]:
                                    total_pert_group_counts[fairness_attribute][str(to_group)] = perturbed_to_group_count
                                else:
                                    total_pert_group_counts[fairness_attribute][str(to_group)] += perturbed_to_group_count
                        
                        # Perturbing the majority groups
                        # While perturbing majority to minority, using correlated_minority of protected attribute for perturbation of correlated attributes
                        if correlation_info is not None:
                            correlation_info["correlated_maj_min"] = correlated_minority

                        for maj in fair_att_inputs["majority"][fairness_attribute]:
                            from_group_count = balanced_pl_sample_total_group_counts[fairness_attribute][str(maj)]
                            to_groups = fair_att_inputs["minority"][fairness_attribute]
                            maj_pert_df, pert_to_groups_count = PerturbationUtils.perform_perturbation(
                                data=balanced_pl_sample_df,
                                column_to_perturb=fairness_attribute,
                                from_group=maj,
                                to_groups=to_groups,
                                spark=self.spark,
                                numerical_perturb_count_per_row=numerical_perturb_count_per_row,
                                float_decimal_place_precision=float_decimal_place_precision,
                                numerical_perturb_seed=numerical_perturb_seed,
                                from_group_count=from_group_count,
                                correlation_info=correlation_info
                            )
                            sample_pert_df = sample_pert_df.unionByName(maj_pert_df) if sample_pert_df is not None else maj_pert_df
                            for to_group in to_groups:
                                perturbed_to_group_count = pert_to_groups_count[str(to_group)]
                                perturbed_data_size += perturbed_to_group_count
                                if str(to_group) not in total_pert_group_counts[fairness_attribute]:
                                    total_pert_group_counts[fairness_attribute][str(to_group)] = perturbed_to_group_count
                                else:
                                    total_pert_group_counts[fairness_attribute][str(to_group)] += perturbed_to_group_count
                        
                        pert_end_time = time.time()
                        self.logger.info("Completed perturbation of the fairness attribute {} for the stratified sample read for balanced dataset calculation. Time taken {} seconds.".format(fairness_attribute, pert_end_time - pert_start_time))
                        
                        # Now scoring the perturbed records
                        # Accumulating the input columns for scoring
                        # Adding "`" around names for columns with `.`
                        for col in input_columns:
                            if "." in col:
                                idx = input_columns.index(col)
                                input_columns[idx] = "`{}`".format(col)
                        sample_pert_df = sample_pert_df[input_columns]
                        # Removing "`" around names for columns with `.`
                        for col in input_columns:
                            if "." in col:
                                idx = input_columns.index(col)
                                input_columns[idx] = col[1:-1]

                        # Scoring the data
                        self.logger.info("Starting the scoring for the perturbed data for fairness attribute {} of the stratified sample for balanced dataset calculation.".format(fairness_attribute))
                        balanced_sample_pert_scoring_start = time.time()

                        predictions_df = scoring_client.score(sample_pert_df, include_features_in_response=True)
                        
                        balanced_sample_pert_scoring_end = time.time()
                        self.logger.info("Completed scoring of the perturbed data of the fairness attribute {} for the stratified sample for the balanced dataset calculation. Time taken {} seconds.".format(fairness_attribute, balanced_sample_pert_scoring_end - balanced_sample_pert_scoring_start))
                        
                        # Appending with the sample payload data read
                        fair_att_balanced_df = balanced_pl_sample_df.unionByName(predictions_df)

                        # Adding the total group counts as well
                        balanced_total_group_counts = BatchUtils.merge_group_counts_dicts({fairness_attribute: balanced_pl_sample_total_group_counts[fairness_attribute]}, total_pert_group_counts)

                        # Calculating the DI on balanced data
                        self.logger.info("Starting the DI calculation of the balanced dataset for fairness attribute {}.".format(fairness_attribute))
                        balanced_data_di_calc_start_time = time.time()

                        # Calculating DI for current fairness attribute on balanced data
                        fair_att_balanced_di_dict = self.calculate_group_bias(fair_att_balanced_df, fair_att_inputs, data_types, model_type, total_group_counts=balanced_total_group_counts)

                        if balanced_di_dict is not None:
                            # Need to add counts for the current fairness attribute in the global balanced DI dict
                            balanced_di_dict[fairness_attribute] = fair_att_balanced_di_dict[fairness_attribute]
                        else:
                            # This is the first DI dict
                            balanced_di_dict = fair_att_balanced_di_dict
                        
                        balanced_data_di_calc_end_time = time.time()
                        self.logger.info("DI calculation of the fairness attribute {} for the balanced dataset completed. Time taken: {} seconds.".format(fairness_attribute, balanced_data_di_calc_end_time - balanced_data_di_calc_start_time))

                    # Adding the perturbed_data_size
                    balanced_di_dict["perturbed_data_size"] = perturbed_data_size
                    
                    balanced_end_time = time.time()
                    balanced_time_taken = balanced_end_time - balanced_start_time
            else:
                if borrow_if_needed and rows_analyzed < min_records and rows_analyzed != 0:
                    # Getting the last processed time
                    last_processed_time = get(
                        monitor_instance, "entity.parameters.last_processed_ts")

                    # Getting the timestamp column name
                    timestamp_column = BatchUtils.get_name_with_modeling_role(
                        TIMESTAMP_MODELING_ROLE, output_data_schema)

                    # Reading the borrowed records
                    borrowed_df = self._read_borrowed_data(
                        subscription, db_name, pl_table_name, min_records - rows_analyzed, timestamp_column, last_processed_time, inputs, schema_name=pl_schema_name)

                    # Calculating DI variables on borrowed records
                    borrowed_di_dict = self.calculate_group_bias(
                        borrowed_df, inputs, data_types, model_type)

                    # Merging the DI values
                    di_dict = BatchUtils.merge_di_dicts(di_dict, borrowed_di_dict)
                elif rows_analyzed == 0:
                    # No new records were read, adding flag to copy the previous measurements, if exists
                    self.logger.info("No new records were read, adding the copy measurement flag in output file.")
                    copy_measurement = True

                payload_end_time = time.time()
                payload_time_taken = payload_end_time - payload_start_time

            output_json = None
            # Checking if enough records were present
            if min_records is not None:
                rows_analyzed = get(di_dict, "rows_analyzed")
                if rows_analyzed < min_records:
                    self.logger.warning("Not enough records received for group bias computation as rows analyzed {} and min_records {}. Hence, not adding the DI dictionary in output file.".format(rows_analyzed, min_records))
                    # Not enough records present in the PL table
                    output_json = {
                        "job_output": [
                            {
                                "data_name": "payload",
                                "rows_analyzed": rows_analyzed,
                                "time_taken": payload_time_taken
                            }
                        ]
                    }

            if output_json is None:
                # Building the output JSON
                self.logger.info("Adding the DI dictionary computed in the output file.")
                output_json = {
                    "job_output": [
                        {
                            "data_name": "payload",
                            "counts": di_dict,
                            "time_taken": payload_time_taken
                        }
                    ]
                }
            self.logger.info("Total time taken for payload dataset calculation was {} seconds.".format(payload_time_taken))

            # Checking if balanced data DI valus are present and adding if yes
            if balanced_di_dict is not None:
                output_json["job_output"].append(
                    {
                        "data_name": "balanced",
                        "counts": balanced_di_dict,
                        "time_taken": balanced_time_taken
                    }
                )
                self.logger.info("Total time taken for balanced dataset calculation was {} seconds.".format(balanced_time_taken))
            
            # Adding the copy measurement flag for `payload` data
            output_json["job_output"][0]["copy_measurement"] = copy_measurement
            
            # Converting the value of outermost value as string because of #19045
            output_json["job_output"] = json.dumps(output_json["job_output"])

            # Write to HDFS
            output_file_name = "{}.json".format(self.name)
            output_path = "{}/{}".format(output_file_path, output_file_name)
            self.save_data(path=output_path, data_json=output_json)
            self.logger.info("The output file successfully stored in HDFS at {}.".format(output_path))
        except Exception as ex:
            error_message = str(ex)
            # If spark job fails with inner join issue which might happen in case of numeric column in older version of spark, provide an informative error. WI #27915
            cartesian_prod_err_phrases = ["cartesian product", "inner join", "join condition is missing", "cross join"]
            if any(phrase.lower() in error_message.lower() for phrase in cartesian_prod_err_phrases):
                error_message = "Please upgrade the Spark version to 3.2 or later as Fairness job failed with error: {}".format(error_message)
            self.save_exception_trace(error_message)
            raise ex
        
        return

    def _read_data(self, subscription: dict, monitor_instance: dict, db_name: str, table_name: str, schema_name: str=None): # (DataFrame, bool)
        """
        Reads and returns data frame for group bias computation.
        :subscription: The subscription object.
        :monitor_instance: The monitor instance object.
        :db_name: The database name.
        :table_name: The table name.
        :schema_name: The name of the schema in which the table resides. [Optional]

        :returns: The data and a flag indicating if records are to be borrowed if required. (from pyspark.sql.dataframe import DataFrame)
        """
        df = None
        borrow_if_needed = False

        # Checking if record timestamp column is present in the PL table
        output_data_schema = get(
            subscription, "entity.asset_properties.output_data_schema")
        timestamp_present = BatchUtils.check_if_modeling_role_present(TIMESTAMP_MODELING_ROLE, output_data_schema)
        
        # Getting the record-id column
        record_id_column = BatchUtils.get_name_with_modeling_role(RECORD_ID_MODELING_ROLE, output_data_schema)
        
        # Getting the probability column (Would be None in case of regression models)
        probability_column = BatchUtils.get_name_with_modeling_role(PROBABILITY_MODELING_ROLE, output_data_schema)

        # Getting all the columns
        columns_to_map = [field["name"] for field in get(output_data_schema, "fields")]
        columns_to_filter = [record_id_column]

        # Getting the min records
        min_records = get(monitor_instance, "entity.parameters.min_records")

        # Getting the partition information
        payload_data_source = BatchUtils.get_data_source_from_subscription(subscription, "payload")
        partition_column = get(payload_data_source, "parameters.partition_column")
        num_partitions = get(payload_data_source, "parameters.num_partitions")
        if num_partitions is None:
            spark_settings = self.arguments.get("spark_settings", BatchUtils.get_spark_settings_from_subscription(subscription))
            if spark_settings is not None:
                num_partitions = int(spark_settings.get("max_num_executors", 1)) * int(spark_settings.get("executor_cores", 1))
            else:
                num_partitions = 1

        # Reading the data
        if timestamp_present:
            self.logger.info("Timestamp column present in the payload logger table.")
            # Checking the last processed time
            last_processed_time = get(monitor_instance, "entity.parameters.last_processed_ts")
            
            # Getting the timestamp column name
            timestamp_column = BatchUtils.get_name_with_modeling_role(
                TIMESTAMP_MODELING_ROLE, output_data_schema)
            
            if min_records is not None:
                self.logger.info("Min records is given in the monitor instance with value {}.".format(min_records))
                if last_processed_time is not None:
                    # This is not the first run, reading from last processed time
                    self.logger.info("This is not the first run for the subscription.")
                    df = DbUtils.get_table_as_dataframe(
                        self.spark,
                        self.location_type,
                        db_name,
                        table_name,
                        schema_name=schema_name,
                        connection_properties=self.jdbc_connection_properties,
                        columns_to_map=columns_to_map,
                        columns_to_filter=columns_to_filter,
                        record_timestamp_column=timestamp_column,
                        start_time=last_processed_time,
                        partition_column=partition_column,
                        num_partitions=num_partitions
                    )
                    borrow_if_needed = True
                else:
                    self.logger.info("This is the first run for the subscription.")
                    # Reading all the records
                    df = DbUtils.get_table_as_dataframe(
                        self.spark,
                        self.location_type,
                        db_name,
                        table_name,
                        schema_name=schema_name,
                        connection_properties=self.jdbc_connection_properties,
                        columns_to_map=columns_to_map,
                        columns_to_filter=columns_to_filter,
                        probability_column=probability_column,
                        partition_column=partition_column,
                        num_partitions=num_partitions
                    )
            else:
                self.logger.info("Min records is not given in the monitor instance.")
                # The `min_records` is not given
                if last_processed_time is not None:
                    self.logger.info("This is not the first run for the subscription.")
                    # This is not the first run, reading from last processed time
                    df = DbUtils.get_table_as_dataframe(
                        self.spark,
                        self.location_type,
                        db_name,
                        table_name,
                        schema_name=schema_name,
                        connection_properties=self.jdbc_connection_properties,
                        columns_to_map=columns_to_map,
                        columns_to_filter=columns_to_filter,
                        record_timestamp_column=timestamp_column,
                        start_time=last_processed_time,
                        partition_column=partition_column,
                        num_partitions=num_partitions
                    )
                else:
                    self.logger.info("This is the first run for the subscription.")
                    # This is the first run, reading all the records
                    df = DbUtils.get_table_as_dataframe(
                        self.spark,
                        self.location_type,
                        db_name,
                        table_name,
                        schema_name=schema_name,
                        connection_properties=self.jdbc_connection_properties,
                        columns_to_map=columns_to_map,
                        columns_to_filter=columns_to_filter,
                        probability_column=probability_column,
                        partition_column=partition_column,
                        num_partitions=num_partitions
                    )
        # This code would be made reachable once #19570 is implemented
        """
        else:
            self.logger.info("Timestamp column not present in the payload logger table.")
            # ------------
            if min_records is None:
                self.logger.info("Min records is not given in the monitor instance.")
                # When both min records and record-timestamp column is not present
                df = DbUtils.get_table_as_dataframe(
                        self.spark,
                        self.location_type,
                        db_name,
                        table_name,
                        schema_name=schema_name,
                        connection_properties=self.jdbc_connection_properties,
                        columns_to_map=columns_to_map,
                        columns_to_filter=columns_to_filter,
                        probability_column=probability_column,
                        partition_column=partition_column,
                        num_partitions=num_partitions
                    )
            else:
                self.logger.info("Min records is given in the monitor instance without timestamp column, which is not supported.")
                # This case is handled at the configuration level,
                # we throw an error in this case
                pass
            # ------------
        """
        return df, borrow_if_needed

    def _read_borrowed_data(self, subscription: dict, db_name: str, table_name: str, num_records: int, timestamp_column: str, last_processed_time: str, inputs: dict, schema_name: str=None): # DataFrame
        """
        Reads and returns the latest borrowed records older than last processed time.
        :subscription: Yhe subscription object.
        :db_name: The database name.
        :table_name: The table name.
        :num_records: The number of records to be read.
        :timestamp_column: The timestamp column in the table.
        :last_processed_time: The last processed time for fairness.
        :inputs: The inputs dictionary.
        :schema_name: The name of the schema in which the table resides. [Optional]

        :returns: Returns the borrowed records data frame. (from pyspark.sql.dataframe import DataFrame)
        """
        df = None

        self.logger.info("Borrowing records from previous window.")

        # Checking if record timestamp column is present in the PL table
        output_data_schema = get(
            subscription, "entity.asset_properties.output_data_schema")
        timestamp_present = BatchUtils.check_if_modeling_role_present(
            TIMESTAMP_MODELING_ROLE, output_data_schema)
        
        # Getting the record-id column
        record_id_column = BatchUtils.get_name_with_modeling_role(RECORD_ID_MODELING_ROLE, output_data_schema)
        
        # Getting the probability column (Would be None in case of regression models)
        probability_column = BatchUtils.get_name_with_modeling_role(PROBABILITY_MODELING_ROLE, output_data_schema)

        # Getting all the columns
        columns_to_map = [field["name"] for field in get(output_data_schema, "fields")]
        columns_to_filter = [record_id_column]

        # Getting the partition information
        payload_data_source = BatchUtils.get_data_source_from_subscription(subscription, "payload")
        partition_column = get(payload_data_source, "parameters.partition_column")
        num_partitions = get(payload_data_source, "parameters.num_partitions")
        if num_partitions is None:
            spark_settings = self.arguments.get("spark_settings", BatchUtils.get_spark_settings_from_subscription(subscription))
            if spark_settings is not None:
                num_partitions = int(spark_settings.get("max_num_executors", 1)) * int(spark_settings.get("executor_cores", 1))
            else:
                num_partitions = 1

        # Reading the data
        df = DbUtils.get_table_as_dataframe(
            self.spark,
            self.location_type,
            db_name,
            table_name,
            schema_name=schema_name,
            connection_properties=self.jdbc_connection_properties,
            columns_to_map=columns_to_map,
            columns_to_filter=columns_to_filter,
            record_timestamp_column=timestamp_column,
            end_time=last_processed_time,
            probability_column=probability_column,
            order_by_timestamp_desc=True,
            partition_column=partition_column,
            num_partitions=num_partitions
        )

        # Getting all labels
        all_labels = get(inputs, "favourable_class") + get(inputs, "unfavourable_class")
        label_column = get(inputs, "class_label")

        # Filtering unused label rows
        model_type = get(subscription, "entity.asset.problem_type")
        if model_type == REGRESSION_MODEL_TYPE:
            all_label_query = SQLUtils.get_num_filter_query(label_column, all_labels)
        else:
            all_label_query = SQLUtils.get_cat_filter_query(label_column, "==", all_labels)
        
        # Applying the query
        df = df.filter(all_label_query)

        # Now applying the limit to get the required number of records
        df = df.limit(num_records)

        return df
